/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef GDSPLOT_HH
#define GDSPLOT_HH

#include <string>
#include "tri_state.hh"
#include "Time.hh"
#include <vector>

class TCanvas;
class TSeries;
class DVector;
class TH1;
class WelchPSD;
class marker;

namespace containers {
    class fSeries;
    class ASD;
    class PSD;
}

/**  gdsplot is a plotting package for gds library objects. The initial 
  *  implementation uses the Cern root package for all graphics, although
  *  this can (and might) be implemented using other (hopefully less 
  *  cumbersome) public domain packages.
  *  \author John Zweizig <john.zweizig@ligo.org>
  *  \version $Id$
  */
class GDSPlot {
public:
    /**  Construct a GDSPlot instance. The GDSPlot constructor is passed a 
      *  optional pointer to the plot unit used by the underlying plotting 
      *  package (Root). If no plot device is specified, One is constructed 
      *  by GDS for this GDSPlot instance.
      *  \brief Construct a new GDSPlot instance.
      *  \param tc Pinter to the Root TCanvas instance to be used.
      */
    GDSPlot(TCanvas* tc=0);

    /**  Destroy this GDSPlot instance. 
      *  \brief Destructor.
      */
    ~GDSPlot(void);

    /**  Plot specified colored boxes.
      *  \brief Plot colored boxes.
      *  \param nbox Number of boxes to be plotted.
      *  \param x0   Vector of minimum x values for each box.
      *  \param x1   Vector of maximum x values for each box.
      *  \param y0   Vector of minimum y values for each box.
      *  \param y1   Vector of maximum y values for each box.
      *  \param z    Vector z values (colors) for each box.
      */ 
    void boxes(size_t nbox, const double* x0, const double* x1, 
	       const double* y0, const double* y1, const double* z);
    
    /**  Histogram contents of a gds DVector. The histogram bins are by 
      *  default nbin (100 0r as specified by set_nbins()) uniformly spaced 
      *  between the xmin and xmax values specified by set_range(). If the
      *  bin limits are not set explicitly or if xmin >= xmax, the limits 
      *  are calculated from the range ov values of the data vector. 
      *  logarithmicly spaced bins may be requested by calling logbin(true);
      *  \brief Histogram contents of a double vector.
      *  \param N     Number of elements to be histogrammed.
      *  \param dv    double vector containing the data
      *  \param title Optional title for the histogram. If \a title is not 
      *               specified, the titl specified by set_title() is used.
      */
    void histogram(const DVector& dv, const std::string& title="");
    
    /**  Histogram contents o double array. The behavior is identical to that
      *  of the DVector histogrammer above.
      *  \brief Histogram contents of a double vector.
      *  \param N     Number of elements to be histogrammed.
      *  \param dv    double vector containing the data
      *  \param title Optional title for the histogram. If \a title is not 
      *               specified, the titl specified by set_title() is used.
      */
    void histogram(size_t N, const double dv[], const std::string& title="");
    
    /**  Build a legend from existing traces. The legend is plotted at the 
      *  position indicated by \a pos.
      *  \brief Build a legend.
      *  \param pos Legend position composed of 'l' (left), 'c' (x center),
      *             'r' (right), 't' (top), 'm' (y middle) or 'b' (bottom).
      */
    void legend(const std::string& pos="br");

    /**  Plot a line from {x1, y1} to {x2, y2}. The coordinates are given in 
      *  plot coordinate system and the line color is as specified by 
      *  set_color().
      *  \param x1 x-coordinate of the start of the line.
      *  \param y1 y-coordinate of the start of the line.
      *  \param x2 x-coordinate of the end of the line.
      *  \param y2 y-coordinate of the end of the line.
      */
    void line(double x1, double y1, double x2, double y2);

    /**  Use logarithmic bins in a histogram.
      *  \brief Logarithmic histogram bins,
      *  \param lb Set logbin parameter.
      */
    void logbin(bool lb);
    
    /**  Clear a canvas and prepare for a new plot. All setting are reset 
      *  to their default values.
      *  \brief Clear canvas.
      */
    void new_plot(void);

    /**  Plot the amplitude spectral density spectrum specified by an ASD
      *  instance.
      *  \brief Plot an amplitude spectrum
      *  \param asd  ASD instance to be plotted.
      */
    void plot(const containers::ASD& asd);

    /**  Plot a specified frequency series instance. If the frequency axis 
      *  has not been otherwise specified it is set to the range of the 
      *  frequency series and is plotted on a logarithmic scale if the 
      *  minimum frequency is positive or zero and the maximum frequency is 
      *  greater than 30 times the minimum frequency. If the minimum frequency
      *  is zero or negative for a log plot, the minimum frequency is set to
      *  the frequency series increment. Similar rules are used to determine
      *  the default scale for the y-axis.
      *  \brief Plot a frequency series
      *  \param pfs  frequency series to be plotted.
      */
    void plot(const containers::fSeries& pfs);

    /**  Plot a Spectrum (PSD) instance.
      *  \brief Plot a spectrum
      *  \param psd  PSD to be plotted.
      */
    void plot(const containers::PSD& psd);
    
    /**  Plot a time-series instance.
      *  \brief plot a time series
      *  \param ts Time series (TSeries) instance to be plotted.
      */
    void plot(const TSeries& ts);
    
    /**  Make a 2D plot of a data vector.
      *  \brief Plot a data vector.
      *  \param dvec  Data vector (DVector) instance to be plotted.
      *  \param xmin  x value of first bin.
      *  \param dx    x value increment
      *  \param title Title string for this plot.
      */
    void plot(const DVector& dvec, double xmin, double dx, 
	      const std::string& title="");

    /**  Write the current canvas contents to the specified file. The file
      *  format is inferred from the extension of the file name. At present
      *  the implemented formats include pdf and png. The ROOT parenthesis
      *  convention can be used to write multiple plots to a file. An open
      *  parenthesis appended to the end if the file name indicates that
      *  output file is to be left open after writing the figure. Further
      *  plots to be written to the same file should specify the only the
      *  file name except the last plot for which a close parenthesis is
      *  appended to the file name.
      *  \brief Print the plot.
      *  \param file File name to which the plot is printed.
      */
    void print(const std::string& file) const;

    /**  Read a file containing a series of colors defined by their RGB values
      *  mapped against a scale value in the range (0, 1). The format of the 
      *  file is 4 columns (scale, red, green, blue) and may contain comments
      *  started with a '#'.
      *  \brief Read a palette file.
      *  \param palette Palette file name.
      */
    void read_palette(const std::string& palette);

    /**  Draw a scatter plot with the specified marker at each specified {x, y}
      *  point.
      *  \brief Draw a scatter plot.
      *  \param N Number of \c \{x,\ y\} points.
      *  \param x x-coordinate of each point.
      *  \param y y-coordinate of each point.
      *  \param requested marker.
      */
    void scatter(size_t N, const double x[], const double y[], const marker& m);

    /**  Draw a scatter plot with the specified marker at each specified {x, y}
      *  point.
      *  \brief Draw a scatter plot.
      *  \param N Number of \c \{x,\ y\} points.
      *  \param x x-coordinate of each point.
      *  \param y y-coordinate of each point.
      *  \param z z (color) coordinate of each point.
      *  \param requested marker.
      */
    void scatter(size_t N, const double x[], const double y[],
		 const double z[], const marker& m);

    /**  Draw a scatter plot with the specified marker at each specified {x, y}
      *  point.
      *  \brief Draw a scatter plot.
      *  \param N Number of \c \{x,\ y\} points.
      *  \param x x-coordinate of each point.
      *  \param y y-coordinate of each point.
      *  \param requested marker.
      */
    void scatter(size_t N, const double x[], const double y[],
		 const marker m[]);

    /**  Set the root TCanvas address.
      *  \brief Set the root TCanvas.
      *  \param tc  Pointer to the canvas to be used.
      *  \param kill_on_exit If set, the canvas will be deleted on destruction.
      */
    void set_canvas(TCanvas* tc, bool kill_on_exit=false);

    /**  Set the color to be used in 1D (histogram, time series, etc.) plots.
      *  \brief Set the trace color.
      *  \param trace_col as defined by root color map.
      */
    void set_color(int trace_col);

    /**  Set the color to be used in 1D (histogram, time series, etc.) plots. 
      *  Index list is in the following table:
      *  <table>
      *  <tr><td>0</td><td>Black</td></tr> 
      *  <tr><td>1</td><td>Blue</td></tr> 
      *  <tr><td>2</td><td>Red</td></tr> 
      *  <tr><td>3</td><td>Green</td></tr> 
      *  <tr><td>4</td><td>Yellow</td></tr> 
      *  <tr><td>5</td><td>Magenta</td></tr> 
      *  <tr><td>6</td><td>Cyan</td></tr> 
      *  <tr><td>7</td><td>Orange</td></tr> 
      *  <tr><td>8</td><td>Spring</td></tr> 
      *  <tr><td>9</td><td>Teal</td></tr> 
      *  <tr><td>10</td><td>Azure</td></tr> 
      *  <tr><td>11</td><td>Violet</td></tr> 
      *  <tr><td>12</td><td>Pink</td></tr> 
      *  </table>
      *  \brief Set trace color by index number.
      *  \param color_index as defined by root color map.
      */
    void set_color_index(int color_index);

    /**  Set the number of histogram bins.
      *  \brief number of histogram bins
      *  \param nb Number of bins.
      */
    void set_nbins(int nb);

    /**  Set the palette to the default color map if it has not already been
      *  set to a user defined map. For now, the default color map is "jet".
      *  \brief Use default color map.
      */
    void set_palette(void);

    /**  Set the palette to a named color map. The currently defined color 
      *  maps are "jet", "hot", "copper", "bone" and viridis.
      *  \brief Select a pre-defined color map
      *  \param palette Color map name.
      */
    void set_palette(const std::string& palette);

    /**  Set a specified parameter to the value(s) represented by the value 
      *  string
      *  \brief Set a parameter/
      *  \param name  Parameter name.
      *  \param value Parameter value string.
      */
    void set_parameter(const std::string& name, const std::string& value);

    /**  Set a specified numeric parameter to the value(s) specified in the 
      *  value vector.
      *  \brief Set a parameter.
      *  \param name  Parameter name.
      *  \param value Parameter numeric value(s).
      */
    void set_parameter(const std::string& name,
		       const std::vector<double>& value);
    
    /**  Set x, and y ranges for following plot(s).
      *  \brief Set x and y range.
      *  \param xmin Minimum x value.
      *  \param xmax Maximum x value.
      *  \param ymin Minimum y value.
      *  \param ymax Maximum y value.
      */
    void set_range(double xmin, double xmax, double ymin, double ymax);

    /**  Set the smoothing. the \a nSmooth argument indicates the number 
      *  of smoothing passes are to be performed. Setting this argument 
      *  zero disables smoothing. The \a smoothOpt argument soecifies the 
      *  kernel to be used. 2D kernels include "k5a" (five points in each
      *  direction, default), "k5b" stronger than k5a, and "k3a".
      *  \brief Set smoothing parameters.
      *  \param nSmooth Number of smoothing passes or zero.
      *  \param smoothOpt Smoothing kernel.
      */
    void set_smooth(int nSmooth, const std::string& smoothOpt);

    /**  Set the statistics box status. If true, a histogram-style statistics 
      *  box is printed. If false the box will not be printed. By defaule the 
      *  statistics box flag is in an nundefined state and will be set 
      *  according to the style.
      *  \brief Set the statistics-box flag.
      *  \param tf If true, a statistics box will be plotted.
      */
    void set_stat_box(bool tf=true);

    /**  Set the graphics style. The default style (0) sets pads to flat white,
      *  disables statistics boxes, etc. Style (1) is used for plots with 
      *  z-axis color-bar and leaves room on the right for the z-axis title. 
      *  This function is used in the constructor.
      *  \brief Set the graphics style.
      *  \param id Style identifier.
      */
    void set_style(int id=0);

    /**  Set the plot size in.
      *  \brief set plot size.
      *  \param xsize X dimension in cm.
      *  \param ysize Y dimension in cm.
      */
    void set_size(double xsize, double ysize) const;

    /**  Set the trace name used in the legend.
      *  \brief Set trace name.
      *  \param name Trace name to be used in the legend;
      */
    void set_trace_name(const std::string& name);

    /**  Set the z range (color dimension) for x-y plots.
      *  \brief Set the z (color) dimension range.
      *  \param zmin Minimum z value.
      *  \param zmax Maximum z value.
      */
    void set_zrange(double zmin, double zmax);

    /**  Plot a spectrogram from the specified time series. If fMin or fMax are
      *  omitted or \e \<=\ 0, they are replaced with \e fMin=1/tSeg and 
      *  \e fMax=f_Nyquist. 
      *  \brief Plot a spectrogram.
      *  \param ts     %Time series
      *  \param tSeg   %Time interval to be plotted.
      *  \param fMin   Minimum frequesncy.
      *  \param fMax   Maximum frequency
      *  \param psdest Estimator to be used to calculate the psd
      */
    void spectrogram(const TSeries& ts, Interval tSeg,
		     double fMin=0, double fMax=0, 
		     const WelchPSD* psdest=0);

    /**  Print current plot parameters.
      *  \brief print GDSPlot status;
      */
    void status(void) const;
    
    /**  Plot a 3D surface using color for the z axis.
      *  \brief Generate a 3d plot.
      *  \param nx  Number of x bins
      *  \param x   Vector of x bin edges (nx+1 elements).
      *  \param ny  Number of y bins
      *  \param y   Vector of y bin edges (ny+1 elements).
      *  \param z   List of nx*ny bin contents with column number (x coordinate)
      *             varying most rapidly, (\e i.e. z[0,0], z[0,1], ... 
      *             z[0,nx-1], z[0,1], ... z[nx-1, ny-1]).
      */
    void surf(int nx, const double* x, int ny, const double* y, 
	      const double* z);

    /**  Set the plot title string.
      *  \brief Plot title string.
      *  \param titl Title string.
      */
    void title(const std::string& titl);

    /**  Return the plot title string.
      *  \brief Plot title string.
      *  \return Title string.
      */
    std::string title(void) const;

    /**  Update the canvas to reflect any changes made since last update;
     */
    void update(void);
    
    /**  Plot a welch PSD estimate of the specified time series. This uses
      *  the default settings, i.e. 1 second span with 50% overlap and Hanning 
      *  window. The fMin and fMax parameters specify the range of frequencies
      *  to be plotted. If fMax is specified/defaulted to zero, the upper
      *  limit is the nyquist frequency. The horizontal and vertical scales 
      *  are set as described for the plot(const fSeries&) method.
      *  \brief Plot estimated PSD.
      *  \param ts     Time series to estimate PSD.
      *  \param fMin   Minimum frequency to plot
      *  \param fMax   Maximum frequency to plot
      *  \param psdest PSD estimator
      */
    void welch(const TSeries& ts, double fMin=0, double fMax=0,
	       const WelchPSD* psdest=0);

    /**  Set the x-axis label string.
      *  \brief X-axis label.
      *  \param titl Text to be plotted below x axis.
      */
    void xlabel(const std::string& titl);

    /**  Set the x axis scale to logarithmic/linear
      *  \brief Set logarithmic x scale.
      *  \param logx set to true for logarithmic or false for linear scale. 
      */
    void xlog(bool logx=true);

    /**  Treat the x axis as a time. Set the units as appropriate to avoid 
      *  excessively large or small numbers.
      *  \brief Set a time axis.
      *  \param dt Time increment.
      *  \param xttl x title to be used.
      *  \return Multiplicative factor used to convert to selected time scale
      */
    double xTimeScale(double dt, const std::string& xttl="Time");
    
    /**  Set the y-axis label string.
      *  \brief Y-axis label.
      *  \param titl Text to be plotted to the left of the y axis.
      */
    void ylabel(const std::string& titl);

    /**  Set the y axis scale to logarithmic/linear
      *  \brief Set logarithmic y scale.
      *  \param logy set to true for logarithmic or false for linear scale. 
      */
    void ylog(bool logy=true);

    /**  Set the z-axis label string.
      *  \brief Z-axis label.
      *  \param titl Text to be plotted to the right of the Z axis color bar.
      */
    void zlabel(const std::string& titl);

    /**  Set the z axis scale to logarithmic/linear
      *  \brief Set logarithmic z scale.
      *  \param logz set to true for logarithmic or false for linear scale. 
      */
    void zlog(bool logz=true);
    
private:
    /** Set the title and axis labels from the stored values.
     */
    void set_axes(TH1& histo);
    void set_pad(void);
    void set_palette(size_t n, const double* x, const double* r,
		     const double* g, const double* b, bool reverse=false);
    void set_range(double& xmin, double& xmax, tri_state logr, size_t N,
		   const double x[]) const;
    void set_zaxis(TH1& histo);
private:
    TCanvas* mCanvas;
    bool     myCanvas;
    int    mTrace;
    double mXmin;
    double mXmax;
    double mYmin;
    double mYmax;
    double mZmin;
    double mZmax;
    tri_state mLogX;
    tri_state mLogY;
    tri_state mLogZ;
    int       mNBin;
    tri_state mLogBin;
    tri_state mStatBox;
    int    mPalletID;
    int    mNextColor;
    int    mNSmooth;
    std::string mTitle;
    std::string mTraceName;
    std::string mXlabel;
    std::string mYlabel;
    std::string mZlabel;
    std::string mSmoothOpt;
};

//======================================  Inline methods
inline std::string
GDSPlot::title(void) const {
    return mTitle;
}

#endif // !defined(GDSPLOT_HH)

