/* -*- mode: c++; c-basic-offset: 3; -*- */
#ifndef GENERATOR_TEXTEVENT_HH
#define GENERATOR_TEXTEVENT_HH

#include "DataSource.hh"
#include <vector>
#include <memory>

namespace generator {

   /**  %TextEvent allows the user to inject pre-recorded data into the 
     *  generation streams. Each %TextEvent source has uses the following 
     *  parameters:
     *  <table>
     *  <tr><td>%Parameter</td>
     *      <td>Contents</td>
     *  </tr>
     *  <tr><td>%Channel</td>
     *      <td>Name of channel providing the data.</td>
     *  </tr>
     *  <tr><td>Files</td>
     *      <td>Path(s) to a json %TSeries streaam file.</td>
     *  </tr>
     *  <tr><td>Mode</td>
     *      <td>Timing mode as described below.</td>
     *  </tr>
     *  <tr><td>%Delay</td>
     *      <td>%Delay from the streamed series start time to the event GPS</td>
     *  </tr>
     *  </table>
     *
     *  The frame data source run in one of four timing modes. These are:
     *  <ol>
     *  <li>\b synch: data are synchronized to the request times. If no data 
     *      are found for the requested time, an exception is thrown.</li>
     *  <li>\b as-available: Data are synchronized to the request times. If  
     *      the requested data are not available they are replaced with 
     *      zeros.</li>
     *  <li>\b delayed: The data are synchronized to the requested time minus
     *      the specified delay.</li>
     *  <li>\b wheel: The first available data from the specified frame list 
     *      are used. When the list of frames runs out, the data restart from 
     *      the beginning of the list.</li>
     *  </ol>
     *
     *  \note The data source sample rate must be the same as 
     *  the sample rate defined in the event data.
     *
     *  Error handing is performed as described for Dacc::fillData(). 
     *  Specifically, the default behavior is to test for NaN, inf or 
     *  unnormalized data in the input series and, if found, to return an 
     *  "Invalid structure data" error exception and in turn, will usually 
     *  cause the calling program (DMTGen) to terminate. This can be avoided
     *  by setting the environment variable DMT_IGNORE_NAN.
     *
     *  \brief Data source reading channels from frame files.
     *  \author John Zweizig
     *  \version 1.1; last modified February 3, 2021
     */
   class TextEvent : public DataSource {
   public:
      using DataSource::setParameter;

   public:
      /** Frame data constructor.
        */
      TextEvent(void);

      /**  Construct a frame data source and initialize the frame file name
        *  and channel name parameters.
        *  @memo Parameter constructor.
        *  @param files   A list of files to be searched for data.
        *  @param channel name of channel to be fetched.
        */
      TextEvent(const char* files, const char* channel);

      /**  Pre-recorded data source destructor.
        *  @memo Destructor.
        */
      virtual ~TextEvent(void);

      /**  Make an identical copy of a frame data source.
        *  @memo Clone a frame data source.
        *  @return Pointer to the newly created object.
        */
      TextEvent* clone(void) const;

      /**  Return the data source type name ("TextEvent").
        *  @memo Source data type name string.
        *  @return Constant pointer to source type name string.
        */
      const char* getDataType(void) const;

      /**  Get the length of the digitized signal for the current event.
        *  @memo Event length
	*  @return Event length in seconds.
	*/
      Interval getDuration(void) const;

      /**  Get the offset of the signal from the event time.
        *  \brief Signal start offset
	*  \return Start time offset in seconds.
	*/
      Interval getStartOffset(void) const;

      /**  Get the time for the next event. This is called when a 
        *  randomization is required for a new event.and may be superceded 
	*  by a the source specific entry.
	*  \brief Next event time
	*  \param current Current time.
	*  \return Time of next event to randomize.
	*/
      Time nextTime(const Time& current);

      /**  Read read events from a json file.
        *  \brief Read one or more events.
	*  \param in istream from which to read json event series.
	*/
      void readEvents(std::istream& in);
      
      /**  Specify the name of the channel to be read in.
        *  @memo Set channel name
        *  @param name Channel name string.
        */
      void setChannel(const std::string& name);

      /**  Set the delay for "delayed" and "wheel" data. The delay is
        *  subtracted from the current (generated) time to give the GPS
        *  to read from the frame.
        *  @memo Set delay time
        *  @param delay Delay time (in seconds)
        */
      void setDelay(const std::string& delay);

      /**  Specify the frame file(s) to be searched for appropriate data. The
        *  frame file(s) may be specified as a single frame path or as a 
        *  globbed path specifier.
        *  @memo Set the file list.
        *  @param frames Frame name list.
        */
      void setFileList(const std::string& frames);

      /**  Specify the timing mode for the frame data. Modes are "synch",
        *  in which the time is requested to match the input data, or 
        *  "wheel" in which the any data is allowed.
        *  @memo Set the timing mode.
        *  @param mode Timing mode name.
        */
      void setMode(const std::string& mode);

      /**  Specify the value of a parameter
        *  @memo Set a parameter 
        *  @param name Parameter name
        *  @param val  Parameter value.
        */
      virtual void setParameter(const std::string& name, 
				const std::string& val);

   private:
      /**  Get the default data type. For frame data, we read in the channel
        *  For the specified epoch, and return the data type from the
        *  frame structure.
        *  \brief Get the channel data type.
        *  \param t0 Start of epoch
        *  \param dt Epoch duration.
        *  \return data type code.
        */
      long defaultType(const Time& t0, Interval dt);

      /**  Get the default sample rate. For frame data, we read in the channel
        *  For the specified epoch, and return the sample rate from the
        *  frame structure.
        *  \brief Get the channel sample rate.
        *  \param t0 Start of epoch
        *  \param dt Epoch duration.
        *  \return Sample rate.
        */
      double defaultRate(const Time& t0, Interval dt);

      /**  Get a data series for the specified time interval.
        *  \brief Get the data time series.
        *  \param t0   Stride start time.
        *  \param dT   Sample step length.
        *  \param N    Number of samples to be read.
        *  \param data Reference to the TSeries into which the data are 
        *              to be placed.
        */
      void getSourceTSeries(const Time& t0, Interval dT, int N, 
			    TSeries& data);

   private:  //-------------------------  private data
      enum timing_mode {
			kSynch,
			kOnce,
			kWheel,
			kDelay
      } mTMode;
      std::string mChannel;
      Interval    mDelay;
      Time        mStartSeg;
      size_t      mEvtIndex;
      std::vector< std::shared_ptr<TSeries> > event_vect;
   };

   //====================================  In-line TextEvent methods.
   inline const char*
   TextEvent::getDataType(void) const {
      return "TextEvent";
   }

   inline Interval
   TextEvent::getStartOffset(void) const {
      return Interval(0.0);
   }

} // namespace generator

#endif  // GENERATOR_TEXTEVENT_HH
